package arbol; 
 

/**
 * Clase que representa un arbol que utiliza la representacion 
 * completamente enlazada, con referencias al padre, hijo izquierdo,
 * y hermano derecho
 */
public class ArbolCE <E> implements Arbol<E>
{
    // Clase interna que representa un nudo del arbol
    private static class Nudo<E> {
        Nudo<E> padre, hijoIzquierdo, hermanoDerecho;
        E contenido;
    }

    // El atributo del arbol es su raiz
    private Nudo<E> raiz;
    
    /**
     * Constructor que crea el arbol vacio
     */
    public ArbolCE ()
        {
            raiz=null;
        }


    /**
     * Constructor al que se le pasa un elemento
     */
    public ArbolCE (E elementoRaiz)
        {
            raiz=new Nudo<E>();
            raiz.contenido=elementoRaiz;
        }

    /**
     * Constructor al que se le pasa un nudo, que sera la raiz del
     * arbol
     */
    ArbolCE (Nudo<E> nudoRaiz)
        {
            raiz=nudoRaiz;
        }
    
    /**
     * Obtener el iterador del arbol
     */
    public IteradorDeArbol<E> iterador() {
        return new IteradorDeArbolCE<E>(this);    
    }

    /**
     * Dejar el arbol vacio
     */
    public void hazNulo() {
        raiz=null;
    }
                
    /**
     * Comprobar si el arbol esta vacio
     */ 
    public boolean estaVacio() {
        return raiz==null;
    }

    /**
     * Clase que representa el iterador del arbol completamente
     * enlazado
     */
    public static class IteradorDeArbolCE<E> implements IteradorDeArbol<E>
    {
        // atributos privados: el arbol y el nudo actual
        private ArbolCE<E> arbol;
        private Nudo<E> nudoActual;

        /**
         * Constructor al que se le pasa el arbol
         * Pone el nudo actual igual a la raiz
         */
        public IteradorDeArbolCE (ArbolCE<E> arbol)
            {
                this.arbol=arbol;
                this.nudoActual=arbol.raiz;
            }

        // Operaciones de modificacion

        /**
         * Anade un hijo al nudo actual, situado mas a la
         * izquierda que los actuales, y con el valor indicado
         */
        public void insertaPrimerHijo(E elemento) throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            Nudo<E> nuevo=new Nudo<E>();
            nuevo.padre=nudoActual;
            nuevo.hermanoDerecho=nudoActual.hijoIzquierdo;
            nuevo.contenido=elemento;
            nudoActual.hijoIzquierdo=nuevo;
        }
    
        /**
         * Anade un hijo al padre del nudo actual, situandolo inmediatamente a 
         * la derecha del nudo actual. Lanza EsRaiz si se intenta anadir 
         * un hermano a la raiz
         */
        public void insertaSiguienteHermano(E elemento) throws EsRaiz, NoValido
            {
                if (nudoActual==null) {
                    throw new NoValido();   
                }
                if (nudoActual.padre==null) {
                    throw new EsRaiz();
                }
                Nudo<E> nuevo=new Nudo<E>();
                nuevo.padre=nudoActual.padre;
                nuevo.hermanoDerecho=nudoActual.hermanoDerecho;
                nuevo.contenido=elemento;
                nudoActual.hermanoDerecho=nuevo;
            }
    
        /**
         * Si el nudo actual es una hoja, la elimina del
         * arbol y hace que el nudo actual sea su padre. Si no es una hoja, 
         * lanza NoEsHoja
         */
        public E eliminaHoja() throws NoEsHoja, NoValido
            {
                if (nudoActual==null) {
                    throw new NoValido();   
                }
                if (nudoActual.hijoIzquierdo!=null) {
                    throw new NoEsHoja();
                }
                ArbolCE<E> rama=(ArbolCE<E>) cortaRama();
                E valor=rama.raiz.contenido;
                return valor;
            }
    
        /**
         * Modifica el contenido del nudo actual reemplazandolo por el 
         * elementoNuevo. Retorna el antiguo contenido del nudo actual
         */
        public E modificaElemento (E elementoNuevo) throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            E valor=nudoActual.contenido;
            nudoActual.contenido=elementoNuevo;
            return valor;
        
        }
    
        /**
         * Elimina la rama del arbol cuya raiz es en nudo actual, y hace 
         * que el nudo actual sea su padre. Retorna la rama cortada como
         * un arbol independiente
         */
        public Arbol<E> cortaRama() throws NoValido
        {
            // comprobar errores
            if (nudoActual==null) {
                throw new NoValido();   
            }

            Nudo<E> padre=nudoActual.padre;
            // si la hoja es la raiz del arbol, lo dejamos vacio
            if (padre==null) {
                ArbolCE<E> rama=new ArbolCE<E>(nudoActual);
                arbol.raiz=null;
                return rama;
            } else {
                // comprobar si la hoja es el hijo izquierdo
                if (padre.hijoIzquierdo==nudoActual) {
                    // es hijo izquierdo
                    padre.hijoIzquierdo=nudoActual.hermanoDerecho;
                } else {
                    // no es hijo izquierdo; 
                    // hay que buscar el hermano izquierdo, n
                    Nudo<E> n=padre.hijoIzquierdo;
                    boolean encontrado=false;
                    while (! encontrado) {
                        if (n.hermanoDerecho==nudoActual) {
                            encontrado=true;
                        } else {
                            n=n.hermanoDerecho;   
                        }
                    }
                    n.hermanoDerecho=nudoActual.hermanoDerecho;
                }
                // preparar el arbol que vamos a devolver
                ArbolCE<E> rama=new ArbolCE<E>(nudoActual);
                rama.raiz.padre=null;
                rama.raiz.hermanoDerecho=null;
                nudoActual=padre;
                return rama;
            }
        }
    
        /**
         * Reemplaza la rama del arbol cuya raiz es el nudo actual, 
         * sustituyendola por nuevaRama; la posicion actual no cambia, 
         * y sera por tanto la raiz de nuevaRama en el arbol actual. 
         * Retorna la rama que ha sido reemplazada como un 
         * arbol independiente.
         */
        public Arbol<E> reemplazaRama(Arbol<E> nuevaRama) throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            // preparar la vieja rama
            Nudo<E> copiaRaiz=new Nudo<E>();
            copiaRaiz.contenido=nudoActual.contenido;
            copiaRaiz.padre=null;
            copiaRaiz.hijoIzquierdo=nudoActual.hijoIzquierdo;
            copiaRaiz.hermanoDerecho=null; 
            ArbolCE<E> viejaRama=new ArbolCE<E>(copiaRaiz);
        
            // copiar la raiz de la nueva rama
            ArbolCE<E> nr=(ArbolCE<E>) nuevaRama;
            nudoActual.contenido=nr.raiz.contenido;
            nudoActual.hijoIzquierdo=nr.raiz.hijoIzquierdo;
        
            return viejaRama;
        }
    
        /**
         * Anade el arbol indicado por nuevaRama haciendo que su raiz
         * sea hija del nudo actual, situandola a la derecha de los 
         * hijos actuales, si los hay
         */
        public void anadeRama(Arbol<E> nuevaRama) throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            // Comprobar si el nudo actual tiene hijos
            if (nudoActual.hijoIzquierdo==null) {
                // no tiene hijos; anadir la rama como hijo izquierdo
                nudoActual.hijoIzquierdo=((ArbolCE<E>)nuevaRama).raiz;
            } else {
                // Buscar el ultimo hijo del nudo actual
                Nudo<E> n=nudoActual.hijoIzquierdo;
                while (! (n.hermanoDerecho==null)) {
                    n=n.hermanoDerecho;   
                }
                n.hermanoDerecho=((ArbolCE<E>)nuevaRama).raiz;
            }
            ((ArbolCE<E>)nuevaRama).raiz.padre=nudoActual;
        }
    
        // Operaciones de consulta

        /**
         * Retorna el elemento contenido en el nudo actual
         */
        public E contenido() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            return nudoActual.contenido;
        }
    
        /**
         * Hace que el nudo actual sea la 
         * raiz del arbol; valdra no valido si el arbol esta vacio
         */
        public void irARaiz()
        {
            nudoActual=arbol.raiz;
        }
    
        /**
         * Hace que el nudo actual sea el primer hijo del actual; 
         * valdra no valido si el nudo actual no tiene hijos
         */
        public void irAPrimerHijo() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            nudoActual=nudoActual.hijoIzquierdo;   
        }
    
        /**
         * Hace que el nudo actual sea el siguiente hermano del actual; 
         * valdra no valido si el nudo actual no tiene hermanos derechos
         */
        public void irASiguienteHermano() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            nudoActual=nudoActual.hermanoDerecho;   
        }
    
        /**
         * Hace que el nudo actual sea el padre del actual; 
         * valdra no valido si el nudo actual era la raiz
         */
        public void irAPadre() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            nudoActual=nudoActual.padre;   
        }
    
        /**
         * Retorna un booleano que indica si el nudo 
         * actual es una hoja o no (es decir si no tiene hijos)
         */
        public boolean esHoja()  throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            return nudoActual.hijoIzquierdo==null;
        }
    
        /**
         * Retorna un booleano que indica si el nudo actual 
         * es la raiz del arbol
         */
        public boolean esRaiz() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            return nudoActual.padre==null;
        }
    
        /**
         *  Retorna un booleano que indica si el nudo actual es el ultimo 
         *  hijo de su padre (es decir si no tiene hermanos derechos)
         */
        public boolean esUltimoHijo() throws NoValido
        {
            if (nudoActual==null) {
                throw new NoValido();   
            }
            return nudoActual.hermanoDerecho==null;
        }
    
        /**
         * Retorna un booleano que indica si el nudo actual es valido, o no
         */
        public boolean esValido() {
            return nudoActual!=null;
        }
    
        /**
         * Retorna un iterador de arbol que es una copia del actual
         */
        public IteradorDeArbol<E> clone() {
            IteradorDeArbolCE<E> copia = new IteradorDeArbolCE<E>(this.arbol);
            copia.nudoActual=this.nudoActual;
            return copia;
        }
    }
    
    /**
     * Lanzada cuando se intenta hacer una operación no permitida
     * para la raíz.
     */
    public static class EsRaiz extends RuntimeException {}
    
    /**
     * Lanzada por la mayoría de las operaciones cuando 
     * el iterador se encuentra en una posición no válida
     */
    public static class NoValido extends RuntimeException {}
    
    /**
     * Lanzada cuando se trata de realizar sobre un nudo interior del árbol
     * una operación que sólo es admisible si el nodo es una hoja 
     */
    public static class NoEsHoja extends RuntimeException {}
}
